#!/usr/bin/env python3

import argparse
import os
import sys
import logging

import pkg_resources
from Registry import Registry

logging.basicConfig()
l = logging.getLogger("regrippy")
l.setLevel("ERROR")


def first(*args):
    for arg in args:
        if arg:
            return arg
    return None


def find_file_nocase(root, file_name):
    for file in os.listdir(root):
        if file.lower() == file_name.lower():
            return file
    return None


def find_path_nocase(root_path, rest_list):
    full_path_parts = []
    for part in rest_list:
        actual_name = find_file_nocase(os.path.join(root_path, *full_path_parts), part)
        if not actual_name:
            return None
        full_path_parts.append(actual_name)
    
    return os.path.join(root_path, *full_path_parts)


def get_hive_paths(args, hive_name):
    if hive_name.lower() in ["all", "ntuser.dat", "usrclass.dat"] and args.all_user_hives and not args.root:
        print("Error: --all-user-hives requires --root", file=sys.stderr)
        sys.exit(3)

    # Basic cases
    if hive_name.lower() == "system":
        path = first(args.system, find_path_nocase(args.root, ["windows", "system32", "config", "system"]) if args.root else None, os.getenv("REG_SYSTEM"))
        return [path] if path else None
    elif hive_name.lower() == "software":
        path = first(args.software, find_path_nocase(args.root, ["windows", "system32", "config", "software"]) if args.root else None, os.getenv("REG_SOFTWARE"))
        return [path] if path else None
    elif hive_name.lower() == "sam":
        path = first(args.sam, find_path_nocase(args.root, ["windows", "system32", "config", "sam"]) if args.root else None, os.getenv("REG_SAM"))
        return [path] if path else None
    elif hive_name.lower() == "ntuser.dat" and not args.all_user_hives:
        path = first(args.ntuser, os.getenv("REG_NTUSER"))
        return [path] if path else None
    elif hive_name.lower() == "usrclass.dat" and not args.all_user_hives:
        path = first(args.usrclass, os.getenv("REG_USRCLASS"))
        return [path] if path else None

    # All user hives
    elif hive_name.lower() in ["ntuser.dat", "usrclass.dat"] and args.all_user_hives:
        hive_paths = []
        users_folder = find_path_nocase(args.root, ["users"])
        if users_folder is None:
            # Not found, let's try "Documents and Settings"
            users_folder = find_path_nocase(args.root, ["Documents And Settings"])
        
        if users_folder is None:
            # Still nothing, we didn't find the Users folder, crash&burn
            raise RuntimeError("Could not find the Users folder")
        
        for user_dir in os.listdir(users_folder):
            if not os.path.isdir(os.path.join(users_folder, user_dir)):
                continue
            if hive_name.lower() == "ntuser.dat":
                path = find_path_nocase(os.path.join(users_folder, user_dir), ["ntuser.dat"])
                if path is not None:
                    hive_paths.append(path)
            else:
                path = find_path_nocase(users_folder, ["appdata", "local", "microsoft", "windows", "usrclass.dat"])
                if path is not None:
                    hive_paths.append(path)
        
        return hive_paths
    
    elif hive_name.lower() == "all":
        hive_paths = []
        for hive in ["system", "software", "sam", "ntuser.dat", "usrclass.dat"]:
            paths = get_hive_paths(args, hive)
            if paths:
                hive_paths.extend(paths)
        return hive_paths
    
    return None


def list_plugins():
    for ep in pkg_resources.iter_entry_points(group='plugins'):
        plugin = ep.load()
        print(f"- {ep.name}({plugin.__REGHIVE__}): {plugin.__doc__}")


def load_plugin(plugin_name):
    for ep in pkg_resources.iter_entry_points(group='plugins'):
        if ep.name == plugin_name:
            return ep.load()

    raise ValueError(f"No such plugin: {plugin_name}")


def main():
    if os.path.basename(sys.argv[0]).lower() != "regrip.py":
        # Issue #5: allow selecting plugins based on argv[0]
        plugin_name = os.path.basename(sys.argv[0])

        # Allow the symlink to be called reg_pluginname to reduce risk of collision
        if plugin_name.startswith("reg_"):
            plugin_name = plugin_name[len("reg_"):]
    else:
        plugin_name = None

    parser = argparse.ArgumentParser(description="Extract information from Windows Registry hives")

    parser.add_argument("--system", "-y", help="Path to the SYSTEM hive. Overrides --root and the REG_SYSTEM environment variable", type=str, default="")
    parser.add_argument("--software", "-o", help="Path to the SOFTWARE hive. Overrides --root and the REG_SOFTWARE environment variable", type=str, default="")
    parser.add_argument("--sam", "-a", help="Path to the SAM hive. Overrides --root and the REG_SAM environment variable", type=str, default="")
    parser.add_argument("--ntuser", "-n", help="Path to the NTUSER.DAT hive. Overrides the REG_NTUSER environment variable", type=str, default="")
    parser.add_argument("--usrclass", "-u", help="Path to the UsrClass.DAT hive. Overrides the REG_USRCLASS environment variable", type=str, default="")
    parser.add_argument("--root", "-r", help="Path to the C: folder.", type=str, default="")
    parser.add_argument("--all-user-hives", help="Work on all NTUSER.DAT and USRCLASS.DAT hives if required. Requires --root. Overrides --ntuser and --usrclass.", action="store_true")
    parser.add_argument("--verbose", "-v", help="Be more verbose", action="store_true")
    parser.add_argument("--pipe", "-p", help="Force output in pipe format", action="store_true")
    parser.add_argument("--list", "-l", help="List available plugins", action="store_true")

    if not plugin_name:
        parser.add_argument("plugin_name", help="Name of the plugin to run", type=str)

    if "--list" in sys.argv or "-l" in sys.argv:
        list_plugins()
        return
    
    args = parser.parse_args()
    if not plugin_name:
        plugin_name = args.plugin_name

    if args.verbose:
        l.setLevel("DEBUG")

    is_pipe = args.pipe or (not os.isatty(sys.stdout.fileno()))
    plugin = load_plugin(plugin_name)

    if type(plugin.__REGHIVE__) == str:
        hive_names = [plugin.__REGHIVE__]
    else:
        hive_names = plugin.__REGHIVE__

    for hive_name in hive_names:
        hive_paths = get_hive_paths(args, hive_name)
        if not hive_paths:
            print("[!] Hive not found:", hive_name, file=sys.stderr)
            continue
        
        for hive_path in hive_paths:
            if hive_path == "-":
                # Special case: read hive from stdin
                reg = Registry.Registry(sys.stdin.buffer)
            else:
                reg = Registry.Registry(hive_path)

            p = plugin(reg, l, hive_name, hive_path)
            results = p.run()

            if results:
                if hive_path != "-" and hive_name == "NTUSER.DAT":
                    p.info(f"User: {p.guess_username()}")
                for result in results:
                    if is_pipe:
                        p.display_machine(result)
                    else:
                        p.display_human(result)


if __name__ == "__main__":
    main()



from regrippy import BasePlugin, PluginResult, mactime


class Plugin(BasePlugin):
    """Lists 'My Recent Documents'"""
    __REGHIVE__ = "NTUSER.DAT"

    def run(self):
        key = self.open_key(r"Software\Microsoft\Windows\CurrentVersion\Explorer\RecentDocs")
        if not key:
            return

        for v in key.values():
            if v.name() == "MRUListEx":
                continue
    
            binary = v.value()
            offset_str_end = binary.find(b"\x00\x00")
            docname = binary[:offset_str_end+1].decode("utf16")

            res = PluginResult(key=key, value=v)
            res.custom['docname'] = docname
            yield res
        
    def display_human(self, result):
        print(result.custom["docname"])
    
    def display_machine(self, result):
        print(mactime(name=f"{self.guess_username()}\t{result.custom['docname']}", mtime=result.mtime))



from regrippy import BasePlugin, PluginResult, mactime
from Registry import Registry


class Plugin(BasePlugin):
    """Lists all services installed on the system"""
    __REGHIVE__ = "SYSTEM"

    def run(self):
        key = self.open_key(self.get_currentcontrolset_path() + "\\Services")
        if not key:
            return
        
        for service in key.subkeys():
            try:
                image_path = service.value("ImagePath").value()
            except Registry.RegistryValueNotFoundException:
                image_path = "N/A"
            
            res = PluginResult(key=service, value=None)
            res.custom["image_path"] = image_path
            yield res
    
    def display_human(self, result):
        print(result.key_name, "//", result.custom["image_path"])
    
    def display_machine(self, result):
        print(mactime(name=f"{result.key_name}\tImagePath={result.custom['image_path']}", mtime=result.mtime))



from regrippy import BasePlugin, PluginResult, mactime


class Plugin(BasePlugin):
    """Gets the name of the last logged-on user"""
    __REGHIVE__ = "SOFTWARE"

    def run(self):
        key = self.open_key("Microsoft\\Windows\\CurrentVersion\\Authentication\\LogonUI")
        if not key:
            return

        for v in key.values():
            if v.name().startswith("LastLoggedOn") and v.name() != "LastLoggedOnProvider":
                res = PluginResult(key=key, value=v)
                yield res
    
    def display_human(self, result):
        print(result.value_name, "\t", result.value_data)
    
    def display_machine(self, result):
        print(mactime(name=f"{result.value_name} {result.value_data}", mtime=result.mtime))



from regrippy import BasePlugin, PluginResult, mactime
from Registry import Registry


class Plugin(BasePlugin):
    """Lists the recently connected to RDP servers"""
    __REGHIVE__ = "NTUSER.DAT"

    def run(self):
        key = self.open_key(r"Software\Microsoft\Terminal Server Client\Servers")
        if not key:
            return
        
        for subkey in key.subkeys():
            try:
                username = subkey.value("UsernameHint").value()
            except Registry.RegistryValueNotFoundException:
                username = ""
            
            res = PluginResult(key=subkey, value=None)
            res.custom['username'] = username
            yield res
    
    def display_human(self, result):
        print(f"{result.custom['username']}@{result.key_name}")
    
    def display_machine(self, result):
        print(mactime(name=f"{self.guess_username()} {result.custom['username']}@{result.key_name}", mtime=result.mtime))



from regrippy import BasePlugin, PluginResult


class Plugin(BasePlugin):
    """Searches for Sysinternal EulaAccepted keys"""
    __REGHIVE__ = "NTUSER.DAT"

    def run(self):
        key = self.open_key(r"Software\Sysinternals")
        if not key:
            return

        for subkey in key.subkeys():
            eula = None
            for v in subkey.values():
                if v.name() == "EulaAccepted":
                    eula = v
                    break

            res = PluginResult(key=subkey, value=eula)
            yield res
    
    def display_human(self, result):
        print(result.key_name, "\t", result.value_data)



import os
from Registry import Registry


def mactime(md5="0", name="", inode=0, mode_as_string="", uid=0, gid=0, size=0, atime=0, mtime=0, ctime=0, btime=0):
    """Formats and returns a Bodyfile-format line.
    All parameters are optional
    """

    return "|".join([md5, name, str(inode), mode_as_string, str(uid), str(gid), str(size), str(atime), str(mtime), str(ctime), str(btime)])


class BasePlugin(object):
    """
    Base class for all plugins. Provides several methods that can be used by plugins
    to perform common actions, like opening registry keys.

    :cvar str_or_list __REGHIVE__: The registry hive (or list thereof) you plugin works on

    :ivar Registry.Registry reg: a handle to a Registry hive
    :ivar logging.Logger logger: a preconfigured Logger. Use the :func:`~BasePlugin.info`, :func:`~BasePlugin.warning` and :func:`~BasePlugin.error` methods instead.
    :ivar str hive_name: the name of the hive. It will always be one of the values in `__REGHIVE__` (see :doc:`createplugin`)
    :ivar str hive_path: the full path to the hive file. Can be "-" if the hive was loaded from `stdin`.
    """

    def __init__(self, reg, logger, hive_name, hive_path):
        self.reg = reg
        self.logger = logger
        self.hive_name = hive_name
        self.hive_path = hive_path

    def run(self):
        """Main entry point of the plugin"""
        raise NotImplementedError("run() needs to be overriden")

    def open_key(self, path):
        """Opens and returns a registry key

        :param path: the full path to a registry key
        :type path: str

        :returns: the key if it was found, otherwise `None`
        :rtype: Registry.RegistryKey
        """
        try:
            key = self.reg.open(path)
            return key
        except Registry.RegistryKeyNotFoundException:
            self.logger.warning("Could not open key " + path)
            return None

    def get_currentcontrolset_path(self):
        """Fetches the path to CurrentControlSet

        :returns: the path to the `CurrentControlSet` key, or `None` if an error happened
        :rtype: str
        """
        select_key = self.open_key("Select")
        if not select_key:
            return None

        current = select_key.value("Current").value()
        ccs_keyname = "ControlSet00" + str(current)

        return ccs_keyname

    def guess_username(self, default=""):
        """Tries to guess the user the current NTUSER.DAT hive corresponds to

        :param default: what to return in case we couldn't determine the user name
        :type default: any type

        :returns: the user name, or `default` if it wasn't found
        :rtype: str
        """

        # Given the following path to a hive
        # /mnt/c/Users/JohnDoe/NTUSER.DAT
        #              ^^^^^^^
        #               we want this part

        dirname = os.path.dirname(self.hive_path)  # /mnt/c/Users/JohnDoe
        folder = os.path.basename(dirname)  # JohnDoe

        if not folder:
            return default
        return folder

    def display_human(self, result):
        """Displays a result to a human. By default,
        it display the path and value of the result.

        :param result: the result to display
        :type result: regrip.PluginResult
        """
        print(result.path, "=>", result.value_name, f"[{result.value_data}]")

    def display_machine(self, result):
        """Displays a result for further processing by
        a machine (piping into mactime for example).

        :param result: the result to display
        :type result: regrip.PluginResult
        """
        print(mactime(name=result.path, mtime=result.mtime))

    def warning(self, msg):
        """Logs a message at WARNING level"""
        self.logger.warning(msg)

    def error(self, msg):
        """Logs a message at ERROR level"""
        self.logger.error(msg)

    def info(self, msg):
        """Logs a message at INFO level"""
        self.logger.info(msg)


class PluginResult(object):
    """A class which holds a single result of a plugin execution

    :ivar dict custom: a `dict` you can use to store custom data for your result
    :ivar int mtime: the "last-modified" time. Automatically set if you pass the `key` parameter.
    :ivar int atime: the "last-accessed" time.
    :ivar int ctime: the "last-changed" time.
    :ivar int btime": the "created" time.
    :ivar str path: complete path to the key. Automatically set if you pass the `key` parameter.
    :ivar str key_name: last part of the key path. Automatically set if you pass the `key` parameter
    :ivar str value_type: the value type
    :ivar str value_name: the name of the Value
    :ivar value_data: the actual value data. The variable type depends on the type of the value.
    """

    def __init__(self, *, key=None, value=None):
        self._key = key
        self._value = value

        self.custom = {}

        self.path = None
        self.mtime = 0
        self.atime = 0
        self.ctime = 0
        self.btime = 0
        self.value_type = None
        self.value_name = None
        self.value_data = None

        if key:
            self.path = key.path()
            self.mtime = int(key.timestamp().timestamp())
            self.key_name = key.name()

        if value:
            self.value_name = value.name()
            self.value_type = value.value_type_str()
            self.value_data = value.value()



from regrippy import BasePlugin, PluginResult, mactime


class Plugin(BasePlugin):
    """Reads OpenSaveMRU and LastVisitedMRU keys (Most Recently Used files in Save As / Open file dialogs)"""
    __REGHIVE__ = "NTUSER.DAT"

    def run(self):
        for path in [r"Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\OpenSaveMRU", r"Software\Microsoft\Windows\CurrentVersion\Explorer\ComDlg32\LastVisitedMRU"]:
            key = self.open_key(path)
            if not key or not key.values():
                continue

            order = key.value("MRUList").value()
            values = dict([(v.name(), v) for v in key.values() if v.name() != "MRUList"])

            for letter in order:
                yield PluginResult(key=key, value=values[letter])
    
    def display_human(self, result):
        print(result.value_data)
    
    def display_machine(self, result):
        print(mactime(name=f"{self.guess_username()}\t{result.value_data}", mtime=result.mtime))



from regrippy import BasePlugin, PluginResult


class Plugin(BasePlugin):
    """Produces a timeline of every key's last modification date"""
    __REGHIVE__ = "ALL"

    def run(self):
        key = self.reg.root()

        yield from self.dump(key)
    
    def dump(self, key):
        res = PluginResult(key=key)
        res.path = self.cleanup_path(res.path)
        yield res

        for subkey in key.subkeys():
            yield from self.dump(subkey)
    
    def cleanup_path(self, s):
        rm = "CsiTool-CreateHive-{00000000-0000-0000-0000-000000000000}"
        if s.startswith(rm):
            return s[len(rm):]
        return s



from regrippy import BasePlugin, PluginResult


class Plugin(BasePlugin):
    """Returns the computer's timezone"""
    __REGHIVE__ = "SYSTEM"

    def run(self):
        ccs = self.get_currentcontrolset_path()
        if not ccs:
            return

        key = self.open_key(ccs + r"\Control\TimeZoneInformation")
        if not key:
            return
        
        value = key.value("TimeZoneKeyName")
        yield PluginResult(key=key, value=value)



#!/usr/bin/env python3

import os
from setuptools import setup

script_folder = os.path.dirname(os.path.realpath(__file__))


def generate_entry_points():
    """
    Walk through plugins folder to generate entry points
    :return: entry points dictionary
    """
    entry_points = {
        'plugins': [],
        'console_scripts': []
    }

    plugins_folder = os.path.join(script_folder, "regrippy", "plugins")
    for plugin_file in os.listdir(plugins_folder):
        if not plugin_file.endswith(".py") or plugin_file == "__init__.py":
            continue
        plugin_name = plugin_file[:-3]
        print(f'add plugin {plugin_name}')
        entry_points['plugins'].append(f'{plugin_name} = regrippy.plugins.{plugin_name}:Plugin')
        entry_points['console_scripts'].append(f'reg_{plugin_name} = regrip:main')

    return entry_points

setup(
    name='regrippy',
    version="1.0.1",
    description="A modern Python-3-based alternative to RegRipper",
    long_description="""RegRip**py** is a framework for reading and extracting useful forensics data from Windows registry hives. It is an alternative to [RegRipper](https://github.com/keydet89/RegRipper2.8) developed in modern Python 3. It makes use of William Ballenthin's [python-registry](https://github.com/williballenthin/python-registry) to access the raw registry hives.

The goal of this project is to provide a framework for quickly and easily developing your own plugins in an incident response scenario.

By default, the script will look for the various hives by reading the `REG_SYSTEM`, `REG_SOFTWARE`, `REG_SAM`, `REG_NTUSER` and `REG_USRCLASS` environment variables. This allows the analyst to simply `export` these in their current shell session and not have to worry about specifying them every time they invoke the script.
Alternatively, you can use the `--root` switch to specify the path to the root of the `C:` drive. RegRippy will automatically look into the right places depending on which hive each plugin needs.

All plugins should also support both a human-readable and machine-readable output (the [Bodyfile](https://wiki.sleuthkit.org/index.php?title=Body_file) format), allowing easy piping to `mactime` or other tools.
    """,
    long_description_content_type="text/markdown",
    author='Airbus CERT',
    author_email='cert@airbus.com',
    url="https://github.com/airbus-cert/regrippy",
    python_requires=">=3.6",
    packages=[
        'regrippy',
        'regrippy.plugins'
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Environment :: Console",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Utilities",
        "Intended Audience :: Information Technology"
    ],
    scripts=['regrip.py'],
    install_requires=[
        'wheel',
        'python-registry'
    ],
    entry_points=generate_entry_points()
)



from regrippy import BasePlugin, PluginResult, mactime


class Plugin(BasePlugin):
    """Reads 'Map Network Drive MRU' key (Most Recently Used remote drives)"""

    __REGHIVE__ = "NTUSER.DAT"

    def run(self):
        key = self.open_key(r"Software\Microsoft\Windows\CurrentVersion\Explorer\Map Network Drive MRU")
        if not key or not key.values():
            return
        
        order = key.value("MRUList")
        values = dict([(v.name(), v) for v in key.values() if v.name() != "MRUList"])

        for letter in order.value():
            res = PluginResult(key=key, value=values[letter])
            yield res
    
    def display_human(self, result):
        print(result.value_data)
    
    def display_machine(self, result):
        print(mactime(name=f"{self.guess_username()}\t{result.value_data}", mtime=result.mtime))



from regrippy import BasePlugin, PluginResult, mactime


class Plugin(BasePlugin):
    """Extracts URLs typed into Internet Explorer"""
    __REGHIVE__ = "NTUSER.DAT"

    def run(self):
        key = self.open_key(r"Software\Microsoft\Internet Explorer\TypedURLs")
        if not key:
            return
        
        self.info(self.guess_username())
        for v in key.values():
            res = PluginResult(key=key, value=v)
            yield res

    def display_human(self, result):
        print(result.value_data)

    def display_machine(self, result):
        print(mactime(name=f"{self.guess_username()}\t{result.value_data}", mtime=result.mtime))
